@@ -18,6 +18,7 @@ import uuid
 from .._utils._platform import _get_structure, Blob
 from .._utils._utils import (ApiInputs, _with_func_attrs, _column_name_cap, _work_order_schema,
                              DiscoveryIntegrationInput, convert_to_pascal_case)
+from .._utils._constants import TAG_LEVEL
 from ..integration._utils import _timezone_offsets, _upsert_entities_affected_count, _adx_support, _timezone_dst_offsets
 
 logger = logging.getLogger(__name__)
@@ -1344,3 +1345,197 @@ def upsert_event_work_order_id(api_inputs: ApiInputs, event_task_id: uuid.UUID,
         return response_status, response.text
 
     return response_status, response.text
+
+@_with_func_attrs(df_required_columns=['Identifier'])
+def upsert_tags(api_inputs: ApiInputs, df: pandas.DataFrame, tag_level: TAG_LEVEL):
+    """
+    Upsert tags to Site/Device/Sensors as specified by the tag_level argument.
+    
+    Required fields are:
+    - Identifier
+    - Additional columns as TagGroups / Tags
+
+    Parameters
+    ----------
+    api_inputs : ApiInputs
+        Object returned by initialize() function.
+    df : DataFrame
+        List of Devices along with corresponding TagsJson to upsert
+    tag_level : TAG_LEVEL
+        Level of tagging applied to the list of Identifier input.
+            If tag_level is Site, Identifier should be InstallationIds.
+            If tag_level is Device, Identifier should be DeviceIds.
+            If tag_level is Sensor, Identifier should be ObjectPropertyIds.
+
+    Returns
+    -------
+    List of affected records
+    
+    """
+    data_frame = df.copy()
+
+    # validate inputs
+    if api_inputs.api_base_url == '' or api_inputs.bearer_token == '':
+        logger.error("You must call initialize() before using API.")
+        return pandas.DataFrame()
+
+    # validate df
+    if data_frame.empty:
+        logger.error("Data Frame is empty. Nothing to upsert.")
+        return pandas.DataFrame()
+
+    # validate tag level
+    if not set([tag_level]).issubset(set(TAG_LEVEL.__args__)):
+        logger.error('tag_level parameter must be set to one of the allowed values defined by the '
+                        'TAG_LEVEL literal: %s', TAG_LEVEL.__args__)
+        return pandas.DataFrame()
+
+    required_columns = getattr(upsert_tags, 'df_required_columns')
+    req_cols = ', '.join(required_columns)
+    proposed_columns = data_frame.columns.tolist()
+
+    # check if required columns are present
+    if not set(required_columns).issubset(proposed_columns):
+        logger.exception('Missing required column(s): %s', set(required_columns).difference(proposed_columns))
+        return 'integration.upsert_tags() - data_frame must contain the following columns: ' + req_cols
+
+    slice_columns = set(proposed_columns).difference(set(required_columns))
+
+    if len(list(slice_columns)) > 0:
+        def update_values(row):
+            j_row = row[list(slice_columns)].to_json()
+            return str(j_row)
+
+        data_frame['TagsJson'] = data_frame[list(slice_columns)].assign(
+            **data_frame[list(slice_columns)].select_dtypes(['datetime', 'object']).astype(str)).apply(update_values, axis=1)
+        data_frame = data_frame.drop(columns=list(slice_columns))
+    else:
+        logger.error("Additional Columns aside from Identifier is required to set as TagsJson.")
+        return pandas.DataFrame()
+
+    switcher_tag_level_identifier = {
+        'Site': 'InstallationId',
+        'Device': 'DeviceId',
+        'Sensor': 'ObjectPropertyId'
+    }
+
+    tag_level_identifier_id = switcher_tag_level_identifier.get(tag_level)
+
+    required_columns = ['TagsJson']
+    required_columns.append(tag_level_identifier_id)
+    
+    data_frame.rename(columns = {'Identifier':tag_level_identifier_id}, inplace = True)
+
+    # remove not required columns
+    data_frame = data_frame.loc[:, required_columns]
+
+    # action request
+    headers = api_inputs.api_headers.integration
+
+    switcher_tag_level_uri = {
+        'Site': f"{api_inputs.api_projects_endpoint}/{api_inputs.api_project_id}/installations/upsert-ingestion-tags",
+        'Device': f"{api_inputs.api_projects_endpoint}/{api_inputs.api_project_id}/devices/upsert-ingestion-tags",
+        'Sensor': f"{api_inputs.api_projects_endpoint}/{api_inputs.api_project_id}/sensors/upsert-ingestion-tags",
+    }
+
+    url = switcher_tag_level_uri.get(tag_level)
+
+    logger.info("Sending request: POST %s", url)
+    response = requests.post(url, data=data_frame.to_json(orient='records'), headers=headers)
+    response_status = '{} {}'.format(response.status_code, response.reason)
+
+    logger.info("Response status: %s", response_status)
+
+    if response.status_code != 200:
+        logger.error("API Call was not successful. Response Status: %s. Reason: %s.", response.status_code,
+                        response.reason)
+    elif len(response.text) == 0:
+        logger.error('No data returned for this API call. %s', response.request.url)
+
+    response_df = pandas.read_json(response.text)
+    response_df.columns = _column_name_cap(response_df.columns)
+
+    return response_df
+
+@_with_func_attrs(df_required_columns=['DeviceId'])
+def upsert_device_metadata(api_inputs: ApiInputs, df: pandas.DataFrame):
+    """
+    Upsert metadata on created devices.
+
+    Required fields are:
+    - DeviceId
+    - Additional columns as Metadata
+
+    Parameters
+    ----------
+    api_inputs : ApiInputs
+        Object returned by initialize() function.
+    df : DataFrame
+        List of Devices along with corresponding MetadataJson to upsert
+
+    Returns
+    -------
+    List of affected Devices
+    """
+    data_frame = df.copy()
+
+    # validate inputs
+    if api_inputs.api_base_url == '' or api_inputs.bearer_token == '':
+        logger.error("You must call initialize() before using API.")
+        return pandas.DataFrame()
+
+    # validate df
+    if data_frame.empty:
+        logger.error("Data Frame is empty. Nothing to upsert.")
+        return pandas.DataFrame()
+
+    required_columns = getattr(upsert_device_metadata, 'df_required_columns')
+
+    req_cols = ', '.join(required_columns)
+    proposed_columns = data_frame.columns.tolist()
+
+    # check if required columns are present
+    if not set(required_columns).issubset(proposed_columns):
+        logger.exception('Missing required column(s): %s', set(required_columns).difference(proposed_columns))
+        return 'integration.upsert_device_metadata() - data_frame must contain the following columns: ' + req_cols
+
+    slice_columns = set(proposed_columns).difference(set(required_columns))
+
+    if len(list(slice_columns)) > 0:
+        def update_values(row):
+            j_row = row[list(slice_columns)].to_json()
+            return str(j_row)
+
+        data_frame['MetadataJson'] = data_frame[list(slice_columns)].assign(
+            **data_frame[list(slice_columns)].select_dtypes(['datetime', 'object']).astype(str)).apply(update_values, axis=1)
+        data_frame = data_frame.drop(columns=list(slice_columns))
+    else:
+        logger.error("Additional Columns aside from DeviceId is required to set as MetadataJson.")
+        return pandas.DataFrame()
+
+    required_columns.append('MetadataJson')
+
+    # remove not required columns
+    data_frame = data_frame.loc[:, required_columns]
+
+    # action request
+    headers = api_inputs.api_headers.integration
+
+    url = f"{api_inputs.api_projects_endpoint}/{api_inputs.api_project_id}/devices/upsert-ingestion-metadata"
+
+    logger.info("Sending request: POST %s", url)
+    response = requests.post(url, data=data_frame.to_json(orient='records'), headers=headers)
+    response_status = '{} {}'.format(response.status_code, response.reason)
+
+    logger.info("Response status: %s", response_status)
+
+    if response.status_code != 200:
+        logger.error("API Call was not successful. Response Status: %s. Reason: %s.", response.status_code,
+                        response.reason)
+    elif len(response.text) == 0:
+        logger.error('No data returned for this API call. %s', response.request.url)
+
+    response_df = pandas.read_json(response.text)
+    response_df.columns = _column_name_cap(response_df.columns)
+
+    return response_df
